package bou.amine.apps.readerforselfossv2.repository

import bou.amine.apps.readerforselfossv2.dao.*
import bou.amine.apps.readerforselfossv2.model.NetworkUnavailableException
import bou.amine.apps.readerforselfossv2.rest.SelfossApi
import bou.amine.apps.readerforselfossv2.model.SelfossModel
import bou.amine.apps.readerforselfossv2.service.ApiDetailsService
import bou.amine.apps.readerforselfossv2.utils.*
import com.github.ln_12.library.ConnectivityStatus
import com.russhwolf.settings.Settings
import io.github.aakira.napier.Napier
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch

class Repository(private val api: SelfossApi, private val apiDetails: ApiDetailsService, private val connectivityStatus: ConnectivityStatus, private val db: ReaderForSelfossDB) {
    val settings = Settings()

    var items = ArrayList<SelfossModel.Item>()
    val isConnectionAvailable = connectivityStatus.isNetworkConnected
    var connectionMonitored = false

    var baseUrl = apiDetails.getBaseUrl()
    lateinit var dateUtils: DateUtils

    var displayedItems = ItemType.UNREAD

    var tagFilter: SelfossModel.Tag? = null
    var sourceFilter: SelfossModel.Source? = null
    var searchFilter: String? = null

    var itemsCaching = settings.getBoolean("items_caching", false)
    var offlineOverride = false

    var apiMajorVersion = 0
    var badgeUnread = 0
    set(value) {field = if (value < 0) { 0 } else { value } }
    var badgeAll = 0
    set(value) {field = if (value < 0) { 0 } else { value } }
    var badgeStarred = 0
    set(value) {field = if (value < 0) { 0 } else { value } }

    init {
        // TODO: Dispatchers.IO not available in KMM, an alternative solution should be found
        CoroutineScope(Dispatchers.Main).launch {
            updateApiVersion()
            dateUtils = DateUtils(apiMajorVersion)
            reloadBadges()
        }
    }

    suspend fun getNewerItems(): ArrayList<SelfossModel.Item> {
        // TODO: Use the updatedSince parameter
        var fetchedItems: List<SelfossModel.Item>? = null
        if (isNetworkAvailable()) {
            fetchedItems = api.getItems(
                displayedItems.type,
                settings.getString("prefer_api_items_number", "200").toInt(),
                offset = 0,
                tagFilter?.tag,
                sourceFilter?.id?.toLong(),
                searchFilter,
                null
            )
        } else {
            if (itemsCaching) {
                fetchedItems = getDBItems().filter {
                    displayedItems == ItemType.ALL ||
                            (it.unread && displayedItems == ItemType.UNREAD) ||
                            (it.starred && displayedItems == ItemType.STARRED)
                }.map { it.toView() }
            }
        }

        if (fetchedItems != null) {
            items = ArrayList(fetchedItems)
            sortItems()
        }
        return items
    }

    suspend fun getOlderItems(): ArrayList<SelfossModel.Item> {
        var fetchedItems: List<SelfossModel.Item>? = null
        if (isNetworkAvailable()) {
            val offset = items.size
            fetchedItems = api.getItems(
                displayedItems.type,
                settings.getString("prefer_api_items_number", "200").toInt(),
                offset,
                tagFilter?.tag,
                sourceFilter?.id?.toLong(),
                searchFilter,
                null
            )
        } // When using the db cache, we load everything the first time, so there should be nothing more to load.

        if (fetchedItems != null) {
            items.addAll(fetchedItems)
            sortItems()
        }
        return items
    }

    suspend fun getMaxItemsForBackground(itemType: ItemType): List<SelfossModel.Item>? {
        return if (isNetworkAvailable()) {
            api.getItems(
                itemType.type,
                200,
                0,
                tagFilter?.tag,
                sourceFilter?.id?.toLong(),
                searchFilter,
                null
            )
        } else {
            emptyList()
        }
    }

    private fun sortItems() {
        items.sortByDescending { dateUtils.parseDate(it.datetime) }
    }

    suspend fun reloadBadges(): Boolean {
        var success = false
        if (isNetworkAvailable()) {
            val response = api.stats()
            if (response != null) {
                badgeUnread = response.unread
                badgeAll = response.total
                badgeStarred = response.starred
                success = true
            }
        } else {
            // TODO: do this differently, because it's not efficient
            val dbItems = getDBItems()
            badgeUnread = dbItems.filter { item -> item.unread }.size
            badgeStarred = dbItems.filter { item -> item.starred }.size
            badgeAll = items.size
        }
        return success
    }

    suspend fun getTags(): List<SelfossModel.Tag>? {
        return if (isNetworkAvailable()) {
            api.tags()
        } else {
            getDBTags().map { it.toView() }
        }
    }

    suspend fun getSpouts(): Map<String, SelfossModel.Spout>? {
        return if (isNetworkAvailable()) {
            api.spouts()
        } else {
            throw NetworkUnavailableException()
        }
    }

    suspend fun getSources(): ArrayList<SelfossModel.Source>? {

        return if (isNetworkAvailable()) {
            api.sources()
        } else {
            ArrayList(getDBSources().map { it.toView() })
        }
    }

    suspend fun markAsRead(item: SelfossModel.Item): Boolean {
        val success = markAsReadById(item.id)

        if (success) {
            markAsReadLocally(item)
        }
        return success
    }

    suspend fun markAsReadById(id: Int): Boolean {
        return if (isNetworkAvailable()) {
            api.markAsRead(id.toString())?.isSuccess == true
        } else {
            insertDBAction(id.toString(), read = true)
            true
        }
    }


    suspend fun unmarkAsRead(item: SelfossModel.Item): Boolean {
        val success = unmarkAsReadById(item.id)

        if (success) {
            unmarkAsReadLocally(item)
        }
        return success
    }

    suspend fun unmarkAsReadById(id: Int): Boolean {
        return if (isNetworkAvailable()) {
            api.unmarkAsRead(id.toString())?.isSuccess == true
        } else {
            insertDBAction(id.toString(), unread = true)
            true
        }
    }

    suspend fun starr(item: SelfossModel.Item): Boolean {
        val success = starrById(item.id)

        if (success) {
            starrLocally(item)
        }
        return success
    }

    suspend fun starrById(id: Int): Boolean {
        return if (isNetworkAvailable()) {
            api.starr(id.toString())?.isSuccess == true
        } else {
            insertDBAction(id.toString(), starred = true)
            true
        }
    }

    suspend fun unstarr(item: SelfossModel.Item): Boolean {
        val success = unstarrById(item.id)

        if (success) {
            unstarrLocally(item)
        }
        return success
    }

    suspend fun unstarrById(id: Int): Boolean {
        return if (isNetworkAvailable()) {
            api.unstarr(id.toString())?.isSuccess == true
        } else {
            insertDBAction(id.toString(), starred = true)
            true
        }
    }

    suspend fun markAllAsRead(items: ArrayList<SelfossModel.Item>): Boolean {
        var success = false

        if (isNetworkAvailable() && api.markAllAsRead(items.map { it.id.toString() })?.isSuccess == true) {
            success = true
            for (item in items) {
                markAsReadLocally(item)
            }
        }
        return success
    }

    private fun markAsReadLocally(item: SelfossModel.Item) {
        if (item.unread) {
            item.unread = false
            badgeUnread -= 1
        }

        CoroutineScope(Dispatchers.Main).launch {
            updateDBItem(item)
        }
    }

    private fun unmarkAsReadLocally(item: SelfossModel.Item) {
        if (!item.unread) {
            item.unread = true
            badgeUnread += 1
        }

        CoroutineScope(Dispatchers.Main).launch {
            updateDBItem(item)
        }
    }

    private fun starrLocally(item: SelfossModel.Item) {
        if (!item.starred) {
            item.starred = true
            badgeStarred += 1
        }

        CoroutineScope(Dispatchers.Main).launch {
            updateDBItem(item)
        }
    }

    private fun unstarrLocally(item: SelfossModel.Item) {
        if (item.starred) {
            item.starred = false
            badgeStarred -= 1
        }

        CoroutineScope(Dispatchers.Main).launch {
            updateDBItem(item)
        }
    }

    suspend fun createSource(
        title: String,
        url: String,
        spout: String,
        tags: String,
        filter: String
    ): Boolean {
        var response = false
        if (isNetworkAvailable()) {
            response = api.createSourceForVersion(
                title,
                url,
                spout,
                tags,
                filter,
                apiMajorVersion
            )?.isSuccess == true
        }

        return response
    }

    suspend fun deleteSource(id: Int): Boolean {
        var success = false
        if (isNetworkAvailable()) {
            val response = api.deleteSource(id)
            if (response != null) {
                success = response.isSuccess
            }
        }

        return success
    }

    suspend fun updateRemote(): Boolean =
        isNetworkAvailable() && api.update()?.isSuccess == true

    suspend fun login(): Boolean {
        var result = false
        if (isNetworkAvailable()) {
            try {
                val response = api.login()
                result = response?.isSuccess == true
            } catch (cause: Throwable) {
                Napier.e(cause.stackTraceToString(), tag = "RepositoryImpl.updateRemote")
            }
        }
        return result
    }

    fun refreshLoginInformation(url: String, login: String, password: String,
                                         httpLogin: String, httpPassword: String,
                                         isSelfSignedCert: Boolean) {
        settings.putString("url", url)
        settings.putString("login", login)
        settings.putString("httpUserName", httpLogin)
        settings.putString("password", password)
        settings.putString("httpPassword", httpPassword)
        settings.putBoolean("isSelfSignedCert", isSelfSignedCert)
        baseUrl = url
        api.refreshLoginInformation()
    }

    private suspend fun updateApiVersion() {
        apiMajorVersion = settings.getInt("apiVersionMajor", 0)

        if (isNetworkAvailable()) {
            val fetchedVersion = api.version()
            if (fetchedVersion != null) {
                apiMajorVersion = fetchedVersion.getApiMajorVersion()
                settings.putInt("apiVersionMajor", apiMajorVersion)
            }
        }
    }

    fun isNetworkAvailable() = isConnectionAvailable.value && !offlineOverride

    fun getDBActions(): List<ACTION> =
        db.actionsQueries.actions().executeAsList()

    fun deleteDBAction(action: ACTION) =
        db.actionsQueries.deleteAction(action.id)

    fun getDBTags(): List<TAG> = db.tagsQueries.tags().executeAsList()

    fun getDBSources(): List<SOURCE> = db.sourcesQueries.sources().executeAsList()

    fun resetDBTagsWithData(tagEntities: List<SelfossModel.Tag>) {
        db.tagsQueries.deleteAllTags()

        db.tagsQueries.transaction {
            tagEntities.forEach { tag ->
                db.tagsQueries.insertTag(tag.toEntity())
            }
        }
    }

    fun resetDBSourcesWithData(sources: List<SelfossModel.Source>) {
        db.sourcesQueries.deleteAllSources()

        db.sourcesQueries.transaction {
            sources.forEach { source ->
                db.sourcesQueries.insertSource(source.toEntity())
            }
        }
    }

    private fun insertDBItems(items: List<SelfossModel.Item>) {
        db.itemsQueries.transaction {
            items.forEach { item ->
                db.itemsQueries.insertItem(item.toEntity())
            }
        }
    }

    private fun getDBItems(): List<ITEM> = db.itemsQueries.items().executeAsList()

    private fun insertDBAction(articleid: String, read: Boolean = false, unread: Boolean = false, starred: Boolean = false, unstarred: Boolean = false) =
        db.actionsQueries.insertAction(articleid, read, unread, starred, unstarred)

    private fun updateDBItem(item: SelfossModel.Item) =
        db.itemsQueries.updateItem(item.datetime, item.title.getHtmlDecoded(), item.content, item.unread, item.starred, item.thumbnail, item.icon, item.link, item.sourcetitle, item.tags.joinToString(","), item.id.toString())


    suspend fun tryToCacheItemsAndGetNewOnes(): List<SelfossModel.Item>? {
        try {
            val newItems = getMaxItemsForBackground(ItemType.UNREAD)
            val allItems = getMaxItemsForBackground(ItemType.ALL)
            val starredItems = getMaxItemsForBackground(ItemType.STARRED)
            insertDBItems(newItems.orEmpty() + allItems.orEmpty() + starredItems.orEmpty())
            return newItems
        } catch (e: Throwable) {}
        return emptyList()
    }
}